#include "lattice.h"

float grd2rad = 180.0/M_PI; // conversion grad -> radian
static int debug = 0;

lattice_window::lattice_window(window & parent, int w, int h, int x, int y) :
    coord_window(parent,w,h,x,y,2,2,2,2) // free rand : hor, vert : 2 Pix
{ 
   scptr = NULL;
   ncolors = 100;
   ixstart = 0; ixend = 0;
   iystart = 0; iyend = 0; rand = 0;
   external_colors = NULL; 
}

// draw clipped lines using global clip vectors y_top, y_bot
// lastp : draw last point of line (1) or not (0)
void lattice_window::cline(XPoint a, XPoint b, int lastp) {
  int sx,sy,dx,dy,x,y,visible,ytt,ybb,up,down;
  int ex = 0, ey = 0;
  if (debug) printf(" %d %d %d %d ",a.x,a.y,b.x,b.y);
  
  dx = b.x - a.x; dy = b.y - a.y;
  sx = (dx > 0) ? 1 : ((dx < 0) ? -1 : 0);
  sy = (dy > 0) ? 1 : ((dy < 0) ? -1 : 0);

  x = a.x; y = a.y; 
  dx = abs(dx); dy = abs(dy);

  do { 
    up = (y > y_top[x]); down = (y < y_bot[x]); 
    if (up) ytt = y; if (down) ybb = y;
    visible = up || down; 
    int endp = (x == b.x) && (y == b.y);
    if (endp &&  ! lastp) break;

    if (visible) DrawPoint(x,y); 

    if (dy < dx) {
      if (up) y_top[x] = ytt; if (down) y_bot[x] = ybb; 
      x += sx; ey += dy;  
      if (2*ey > dx) { y += sy; ey -= dx; } 
    } else {
      y += sy; ex += dx;
      if (2*ex > dy) {  
	if (up) y_top[x] = ytt; if (down) y_bot[x] = ybb; 
	x += sx; ex -= dy; }
    }
    if (endp) break; 
  } while (1);  
}

// fills from the line (a,b) up to the boundaries y_top, y_bot
// if (lastp) including the endpoint, 
// if (dofill) really fill, else only set boundaries
void lattice_window::fill(XPoint a, XPoint b, int lastp, int dofill) {
  int sx,sy,dx,dy,x,y;
  int ex = 0, ey = 0;

  dx = b.x - a.x; dy = b.y - a.y;
  sx = (dx > 0) ? 1 : ((dx < 0) ? -1 : 0);
  sy = (dy > 0) ? 1 : ((dy < 0) ? -1 : 0);

  x = a.x; y = a.y; 
  dx = abs(dx); dy = abs(dy);

  if (dy < dx) { 
    do {
      int endp = (x == b.x) && (y == b.y);
      if (endp &&  ! lastp) break;

      if (y > y_top[x]) { if (dofill) line(x,y,x,y_top[x]); y_top[x] = y; }
      if (y < y_bot[x]) { if (dofill) line(x,y,x,y_bot[x]); y_bot[x] = y; }
      x += sx; ey += dy;  
      if (2*ey > dx) { y += sy; ey -= dx; } 
      if (endp) break; 
    } while (1);  
  } else {
    do {
      int endp = (x == b.x) && (y == b.y);
      if (endp &&  ! lastp) break;

      y += sy; ex += dx;
      if (2*ex > dy) {  
	if (y > y_top[x]) { if (dofill) line(x,y,x,y_top[x]); y_top[x] = y; }
        if (y < y_bot[x]) { if (dofill) line(x,y,x,y_bot[x]); y_bot[x] = y; }
	x += sx; ex -= dy; }
    
      if (endp) break; 
    } while (1);  
  }
}

struct XPoint lattice_window::screen_project(float x, float y, float z)
{
  float xs,ys,yss,zss,t,dx,dy; 
  // printf(" %f %f %f \n",x,y,z);
  dx = x - xp; dy = y - yp;
  xs = ca*dx - sa*dy;  ys = ca*dy + sa*dx;     // rotation x,y-plane
  zss = cb*z - sb*ys; yss = cb*ys + sb*z;      // rotation ys,z-plane
    
  // central-projection with perspective
  // exaktly : temp.y = arcsin(yss/(distance - zss)) = y-view angle
  // where "distance" is distance between viewpoint and origin

  t = (dist == 0.0) ? 1.0 : 1.0/(1.0 - dist*zss/xp); 	   
  XPoint temp = p_window(t*xs,t*yss);

  // this should not happen, it would overflow y_top, y_bot indicees
  // clip to boundary values
  if (temp.x < 0) temp.x = 0; else if (temp.x >= width) temp.x = width-1;
  return temp;
}

void reverse(int &x1, int &x2, int &dx) {
  int tmp = x1; x1 = x2; x2 = tmp; dx = -dx;
}

// the old method for lattice drawing is obsolete
// alpha, beta : Drehwinkel (hier rad!), gamma : z-Faktor,
/*
void lattice_window::make_lattice(int nx, int ny, float *FF, 
			       float alpha, float beta, float gamma, float z0, 
			       int opaque, float distance)
{ 
  float (*ff)[ny] = (float (*) [ny]) FF;
  int x,y; 
  y_bot = new int[width]; y_top = new int[width]; 
  for (x=0; x < width; x++) { y_top[x] = 0; y_bot[x] = height; }

  ca = cos(alpha); sa = sin(alpha); cb = cos(beta); sb = sin(beta);
  dist = distance;
  // printf(" %d %d %f %f %f\n",nx,ny,alpha,beta,gamma);

  // the ix-range [xstart .. xend] will be mapped to [0..xdraw-1] 
  int xdraw,ydraw;     // the x,y-lattice values, to be drawn
  if (ixend == 0) ixend = nx; // with 0 initialized
  if (iyend == 0) iyend = ny;
  xdraw = ixend - ixstart + 2; // + 2 points for border
  ydraw = iyend - iystart + 2;
  
  xspan = ixend - ixstart - 1; yspan = iyend - iystart -1;
  xp = 0.5*xspan; yp= 0.5*yspan; // rotation centre in x,y plane
  
  float xmax,ymax;
  ymax = xmax = 0.5*sqrt(xspan*xspan + yspan*yspan); 

  //    printf( "%d %d %d %d \n",ixstart,iystart,xdraw,ydraw);
  define_coord(-xmax, -ymax, xmax, ymax); 

  int ix,iy;
  float zh,xh,yh;
  zmax = -1E30; 

  struct XPoint scp[xdraw][ydraw]; // all grid-points after transformation
  // x,y : indicees of  scp ; 
  // xh,yh,z : 3d-coordinates
  // ix,iy : indicees of  ff 
  // die Ecken zB. (xh,yh,zh) = (0,0,0) -> will be mapped three times to :  
  //                x,y = (0,0) (0,1) (1,0) 
  // xh,yh : (0,0,zh) -> (1,1)
  znorm = gamma * xmax;
  for (x = 1, ix = ixstart, xh = .0; x < xdraw-1; x++, ix++, xh += 1.0) { 
    for (y = 1, iy = iystart, yh = .0; y < ydraw-1; y++, iy++, yh += 1.0 ) {
      zh = znorm * (ff[ix][iy]- z0); if (zh > zmax) zmax = zh; 
      scp[x][y] = screen_project(xh,yh,zh);
    }
  }
  yh = ydraw-3;
  for (x = 1, xh = 0.0; x<xdraw-1; x++, xh += 1.0) { 
    scp[x][0] = screen_project(xh,0,0);
    scp[x][ydraw-1] = screen_project(xh,yh,0); 
  }
  
  xh = xdraw-3;
  for (y = 1, yh = 0.0; y<ydraw-1; y++, yh += 1.0) { 
    scp[0][y] = screen_project(0,yh,0);
    scp[xdraw-1][y] = screen_project(xh,yh,0); 
  }
  scp[0][0] = scp[0][1];  scp[0][ydraw-1] = scp[0][ydraw-2];  
  scp[xdraw-1][0] = scp[xdraw-2][0];  
  scp[xdraw-1][ydraw-1] = scp[xdraw-2][ydraw-2]; 

  int x0,y0,dx,dy,xx,yy,yt,x1,y1;

  XPoint sc00 = screen_project(0,0,0),       // the origin   
         scxy = screen_project(xdraw-3,ydraw-3,0), // the opposite edge       
         scx = screen_project(xdraw-3,0,0),  // x,0
         scy = screen_project(0,ydraw-3,0),  // 0,y
         sczm = screen_project(0,0,zmax);   
  pline(sc00,sczm);       // z-axis
  
  int i, dzy = sczm.y - sc00.y, nti; // draw z-ticks with numbers
  nti = (abs(dzy) > 60) ? 5 : (abs(dzy) / 15 + 1); // number of ticks 
  for (i=0; i< nti+1; i++) { 
    float zi = i/float(nti);
    int x = sc00.x, y = sc00.y + int(zi*dzy); 
    line(x,y,x-4,y); // z tick
    char ticks[20]; sprintf(ticks,"%g",zi*zmax/znorm + z0);
    int xs = (x < width/2) ? x+3 : x - 6*strlen(ticks) - 6; // string position
    DrawString(xs,y+5,ticks);
  }

  // determination of direction of   for-loops xx,yy
  // so, that y is decreasing, ie start always in most forward point
  // rand = 1 : drawing with border lines  , 
  // else  rand = 0 : ie only from x = 1 .. xdraw-2 to draw
  int orand = ! rand; // 0/1
  x0 = orand; x1 = xdraw-orand-1; dx = 1;
  if (sin(alpha) < 0.) reverse(x0,x1,dx);

  y0 = orand; y1 = ydraw-orand-1; dy = 1;
  if (cos(alpha) < 0.) reverse(y0,y1,dy);
 
  for (yy = y0 ; ; yy+= dy) {
      for (xx = x0 ; ; xx+= dx) 
	{ XPoint sc = scp[xx][yy];
	  if (debug) printf("\n %d %d ", xx, yy);
	  if (xx != x1) xline(sc,scp[xx+dx][yy],1);
          if (yy != y1) xline(sc,scp[xx][yy+dy],0); 
	  if (xx == x1) break;
	} 
      if (yy == y1) break;
  }
  // at least draw the xy axes
  if (orand) { cline(sc00,scx,1); cline(sc00,scy,1); } 
  
   delete[] y_top; delete[] y_bot;
}
*/

// for swapping the driving step variables 
static void swap_dir(int& x0, int& x1, int &dx, float &f0, float &f1) {
  int t = x0; x0 = x1; x1 = t; 
  float f = f0; f0 = f1; f1 = f;
  dx = -dx; 
}

palette_popup *pal_win = NULL; // popup window der color palette 

// draws a perspective picture of the gridwise defined function FF
// either with shading algorithm (body = True) or as a lattice
void lattice_window::make_body(int nx, int ny, float *FF, 
			       float alpha, float beta, float gamma, float z0, 
			       int opaque, 
			       float distance, float al, float bl) {
  float (*ff)[ny] = (float (*) [ny]) FF;
  y_bot = new int[width]; y_top = new int[width]; 
  int x,y; 
  for (x=0; x < width; x++) { y_top[x] = 0; y_bot[x] = height; }

  // eigentlich koennen hier alle Projektionen und clippings entfallen !
  if (flat_mode) { alpha = beta = distance = 0; }

  ca = cos(alpha); sa = sin(alpha); cb = cos(beta); sb = sin(beta);
  dist = distance;
  // printf(" %d %d %f %f %f\n",nx,ny,alpha,beta,gamma);

  if (ixend == 0) ixend = nx; // am Anfang mit 0 initialisiert
  if (iyend == 0) iyend = ny;
  xspan = ixend - ixstart - 1; yspan = iyend - iystart -1;
  xp = 0.5*xspan; yp= 0.5*yspan; // Drehpunkt der x,y-Ebene

  float xmax,ymax;
  if (flat_mode) { xmax = xp; ymax= yp; } else
  ymax = xmax = 0.5*sqrt(xspan*xspan + yspan*yspan); 

  define_coord(-xmax, -ymax, xmax, ymax); 

  int ix,iy;
  float zh,xh,yh;
  zmax = -1E30;
  if (scptr) delete[] scptr; 
  scptr = new XPoint[nx*ny]; 
  // all  nx*ny grid-points after trafo, addressed as scp[nx][ny]
  struct XPoint (*scp)[ny] = (XPoint (*) [ny]) scptr; 
  znorm = gamma * xmax;
  for (ix = ixstart, xh = .0; ix < ixend; ix++, xh += 1.0) { 
    for (iy = iystart, yh = .0; iy < iyend; iy++, yh += 1.0 ) {
      zh = znorm * (ff[ix][iy]-z0); if (zh > zmax) zmax = zh;
      if (flat_mode) { 
	scp[ix][iy] = p_window(xh - xp,yh - yp);
      } else
      scp[ix][iy] = screen_project(xh, yh, zh);
    }
  }  

  int colors[nx][ny];

  if (body && ! external_colors) { 
    // else : given colors for areas -> no computation               
    // computing of reflections coefficients of the grid points
    // subtraction of rotation angles -> means fixed (constant) light source
    // otherwise : co-moving light source (moves with frame)
    // disadvantage for fixed light source :
    // upon rotating the orientation is bad 
    float e_alpha = al /* -alpha */ , e_beta = bl /* -beta */ ;
    float ex = cos(e_alpha)*cos(e_beta), ey = sin(e_alpha)*cos(e_beta), 
    ez = sin(e_beta); // (ex,ey,ez) = normal of incident ray

    float rpp[nx][ny];
    float rmax = -1e30, rmin = 1e30, mu = 1/SQR(gamma); 
    for (ix = ixstart+1; ix < ixend; ix++) { 
      for (iy = iystart+1; iy < iyend; iy++) {
	float z00 = ff[ix][iy]; // z0 shift not needed here
	float z10 = ff[ix-1][iy], z01 = ff[ix][iy-1], z11 = ff[ix-1][iy-1];
	// computation of normals  to (x,y) (x+1,y) (x,y+1) :
	// n = (zx, zy, 1)/|N|
	float zxp = z01 - z11,  zyp = z10 - z11; // growing differences
	float zxm = z00 - z10,  zym = z00 - z01; // falling  differences
	// fuer Quadrate benutze mittlere Ableitung
         zxp += zxm; zyp += zym; 
	//            scalar product n * E
	rpp[ix][iy]= opaque ? (zxp*ex+zyp*ey+ez)/sqrt(zxp*zxp + zyp*zyp + mu) 
	                    : - (z00+z01+z10+z11); // prop. z
	rmax = MAX(rpp[ix][iy], rmax);  // for norming of scalar products
	rmin = MIN(rpp[ix][iy], rmin);
      }
    } 
    // printf(" %g %g\n",rmin,rmax);
    for (ix = ixstart+1; ix < ixend; ix++)  
      for (iy = iystart+1; iy < iyend; iy++)
	// norming  of scalar prod. s = [rmin..rmax] -> [0..ncolors-1]
	colors[ix][iy] = (int) ((ncolors-1)*(rpp[ix][iy]-rmin)/(rmax-rmin));
        // norming in -1..+1 (for normals adapted) :
	// colors[ix][iy] = (int) ((ncolors-1)*(0.5*rpp[ix][iy] + 0.5));
	
  }

  // Determination of parameters of the for-loops, in such a way that 
  // the drawing always starts at the front edge (towards view-point)        
  // xs,ys  are the offsets between  ix,iy and the 
  // index of the scalarprodukt-arrays, which count from 1,1
  // examples  : ixstart = 0, ixend = 10, 
  //                	x0   	x1    xe     dx     xs
  // alpha = 0..90   	0    	1      9      1      0
  // alpha = -90..0  	9    	8      0     -1      1
  
  int x0 = ixstart, x1 = ixstart+1, xe = ixend-1, dx = 1, 
      xx = ixend-ixstart-1, xs = 0;  
  int y0 = iystart, y1 = iystart+1, ye = iyend-1, dy = 1, 
      yy = iyend-iystart-1, ys = 0;

  float xfront = 0, yfront = 0, xback = xspan, yback = yspan, xff, yff;
  if (sa < 0.) { swap_dir(x0,xe,dx,xfront,xback); x1 = ixend-2; xs = 1; }
  if (ca < 0.) { swap_dir(y0,ye,dy,yfront,yback); y1 = iyend-2; ys = 1; }

  // draw the x and y axes (front lines)
  XPoint sff = screen_project(xfront,yfront,0),
         sbf = screen_project(xback,yfront,0), 
         sfb = screen_project(xfront,yback,0);
  cline(sff,sbf,0); // x axis
  cline(sff,sfb,0); // y axis

  if (rand) { // draw the lines from axes to first grid line
    for (y = y0, yff = yfront; ; y += dy, yff += dy) {
      cline(screen_project(xfront,yff,0),scp[x0][y],0);
      if (y == ye) break;
    }
    for (x = x0, xff = xfront; ; x+= dx, xff += dx) {
      cline(screen_project(xff,yfront,0),scp[x][y0],0);  
      if (x == xe) break;
    }
  }

  if (body) { // draw coloured  faces 
    for (ix=x1, x=0; x < xx; ix+=dx, x++) 
      cline(scp[ix][y0],scp[ix-dx][y0],1);
    // or : fill(scp[ix][y0],scp[ix-dx][y0],1,0); -> without drawing
    for (iy=y1, y=0; y < yy; iy+=dy, y++)
      cline(scp[x0][iy],scp[x0][iy-dy],1);

    for (ix=x1, x=0; x < xx; ix+=dx, x++) {
      for (iy=y1, y=0; y < yy; iy+=dy, y++) { 
	XPoint p00 = scp[ix][iy], p10 = scp[ix-dx][iy], p01 = scp[ix][iy-dy];

	int color;
	if (external_colors) color = external_colors[(ix+xs)*ny + iy+ys];
	else {
	  int pixr = colors[ix+xs][iy+ys];
	  color = pal_win->color_cells[pixr].pixel;
	} 
	set_color(color);  
	fill(p00, p10, 0, 1); fill(p00, p01, 0, 1);  
      }
    }
    set_color(black); // reset for gc_copy !  
  } else {  // usual lattice is drawn
    for (y = y0; ; y += dy) {
      for (x = x0; ; x+= dx) { 
	XPoint sc = scp[x][y];
	if (x != xe) xline(sc,scp[x+dx][y],1);
	if (y != ye) xline(sc,scp[x][y+dy],0); 
	if (x == xe) break;
      } 
      if (y == ye) break;
    }
  }
  if (! flat_mode) {
    XPoint sc00 = screen_project(0,0,0);       // the origin
    XPoint sczm = screen_project(0,0,zmax); 
    pline(sc00,sczm);     // z-axis to zmax
    int i, dzy = sczm.y - sc00.y, nti;  // draw z-ticks with numbers
    nti = (abs(dzy) > 60) ? 5 : (abs(dzy) / 15 + 1); // number of ticks 
    for (i=0; i< nti+1; i++) { 
      float zi = i/float(nti);
      int x = sc00.x, y = sc00.y + int(zi*dzy); 
      line(x,y,x-4,y); // z tick
      char ticks[20]; sprintf(ticks,"%g",zi*zmax/znorm + z0);
      int xs = (x < width/2) ? x+3 : x - 6*strlen(ticks) - 6; //string Position
      DrawString(xs,y+5,ticks);
    }
  }
  delete[] y_top; delete[] y_bot;
}

